#ifndef general_h
#define general_h

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <math.h>

#define FALSE 0
#define TRUE 1

#define NUM_TEAMS 16

/* Visible field blocks.                         */
/* Try to keep MAXX : MAXY = 4 : 3 aspect ratio. */
#define MAXX 16			/* was 20 */
#define MAXY 12			/* was 15 */
#define MINX 0
#define MINY 0

/* Complete battlefiled dimensions. Typically four quadrants each
   MAXX*MAXY in area. */

#define FIELDX (MAXX+MAXX)
#define FIELDY (MAXY+MAXY)

/* Always maintain a compass ordering of these constants. */
#define FULL_ROTATION 4
#define NORTH  0
#define EAST   1
#define SOUTH  2
#define WEST   3


#define NORTHEAST 4
#define SOUTHEAST 5
#define SOUTHWEST 6
#define NORTHWEST 7

#define MAX_ROBOTS 10

#define MAX_TASKS 3
#define MISSILE_VELOCITY 5 /* Number cells / per unit time. */
#define MISSILE_DESTRUCTION 5
#define UNCHARTED -500
#define MAXDIST 5000
#define NOT_IN_WORLD -1

#define NUM_OBJECT_TYPES 7

/* Note: must maintain a 1-to-1 correspondence w.r.t. compass points above. */
#define N_MISSILE 0
#define E_MISSILE 1
#define S_MISSILE 2
#define W_MISSILE 3

#define MAX_MISSILES ((2*MAXX)+(2*MAXY))

#define FRIENDLY_ROBOT 4
#define ENEMY_ROBOT    5
#define MISSILE        6
#define POS_OBS        7
#define NEG_OBS        8
#define STATIC_OBS     9
#define UNKNOWN        10

#define OBJECT_OFFSET 4

#define GOAL          30
#define GENERIC_ROBOT 31 

#define UNTIL_BOUNDARY 50

#define WALK 0
#define ROTATE 1
#define FIRE 2

#define CLOCKWISE 100
#define ANTICLOCKWISE 200

#define MAX_RANGE 4
#define INIT_LIFE 5

#define MISSILE_INACTIVE 0
#define MISSILE_ACTIVE 1

#define T_FAIL 0    /* Robot task was not completed succesfully. */
#define T_SUCCESS 1 /* Robot task was successfully completed. */

typedef struct coord
{
  int x;
  int y;
} coordinate;


typedef struct missile_struct
{
  coordinate pos;
  int status;
  int orientation;
} missile_type;


struct robot_args
{
  int robot_num;
  int value;     /* Generic value passable to many different tasks. */
  
  /* The following 'x' and 'y' members are coordinates associated 
     with the robot whose id is 'robot_num'. They are only necessary 
     with non-sequentially-consistent consistency protocols. I feel 
     ashamed at having to add them :-( */
  int x;
  int y;
  int orientation;		/* Orientation of robot. */
};

/* ID of local client. Global since passed around everywhere. */
extern int id;   

extern int robot_front [NUM_TEAMS][MAX_ROBOTS];
extern int robot_life [NUM_TEAMS][MAX_ROBOTS];
extern coordinate robot_position [NUM_TEAMS][MAX_ROBOTS];
extern int killed_neg [NUM_TEAMS][MAX_ROBOTS];
extern missile_type missile[MAX_MISSILES];

extern coordinate goal; /* Same global goal for all robots. */
extern coordinate environment [MAX_ROBOTS][NUM_OBJECT_TYPES+OBJECT_OFFSET];
extern int sighted_flag [MAX_ROBOTS][NUM_OBJECT_TYPES+OBJECT_OFFSET];
extern int world [FIELDX][FIELDY];

void create_task (int sched_num,
		  int robot_id,
		  int sighted_flag[MAX_ROBOTS]
		  [NUM_OBJECT_TYPES+OBJECT_OFFSET],
		  coordinate environment [MAX_ROBOTS]
		  [NUM_OBJECT_TYPES+OBJECT_OFFSET]);

void createTaskFromInput (int sched_num, int robot_id, 
			  int button_number, int rotation);

float distance (coordinate c1, coordinate c2);

int in_world (coordinate cur_pos);

int equal (coordinate tmp1, coordinate tmp2) ;

int already_visited (coordinate c);

coordinate up_left (coordinate tmp);

coordinate up (coordinate tmp);

coordinate up_right (coordinate tmp);

coordinate left (coordinate tmp);

coordinate right (coordinate tmp);

coordinate dn_left (coordinate tmp);

coordinate dn (coordinate tmp);

coordinate dn_right (coordinate tmp);

void init_teams ();

void init_world (int id, int place_obstacles);

int zap_world ();

int rotate_clockwise (struct robot_args *arg);

int rotate_anticlockwise (struct robot_args *arg);

int walk (struct robot_args *arg);

int fire_missile (struct robot_args *arg);

int next_move (coordinate src, coordinate dest);

int calculate_walk_orientation (int robot_id);

void update_missile (int missile_index);
	
void wipeObject (char *sector, int x, int y);

void drawRobot (char *sector, int from_x, int from_y, 
		int to_x, int to_y, int orientation, 
		int type, char *data_buf);

void drawNegObs (char *sector, int x, int y);

void drawPosObs (char *sector, int x, int y);

void drawStaticObs (char *sector, int x, int y);

void drawUnknown (char *sector, int x, int y);

void drawObject (char *sector, int x, int y, 
		 int orientation, int object_type,
		 char *data_buf);

void drawMissile (char *sector, int from_x, int from_y, 
		  int to_x, int to_y, int orientation);

void doExplosion (char *sector, int x, int y);

void glorySequence (char *sector, int x, int y, int type);

void redrawGoal (char *sector, coordinate goal);

void snooze (unsigned long interval);

void getRobotInfo (int x, int y, int *team_number, int *robot_number);

#endif



