#include <math.h>
#include "socklink.h"
#include "consistency.h"

int world [FIELDX][FIELDY];
team_t team [NUM_TEAMS];

textmsg_t rtext, stext;		   /* Receive/send text msg buffers. */
datamsg_t sdata;
static char msg_buf [NET_MAX_PACKET_SIZE];
static int msg_len;
int conn_id;

char *progname;
static int num_robots = 0;         /* Total #robots in game (all clients). */
static int num_clients = 2;
static int registered_clients;     /* Actual #clients signed-on and active. */
static int goal_score;             /* Score for reaching goal. */
coordinate goal;                   /* Same global goal for all robots. */
static int place_obstacles = TRUE;
static int static_obstacles = FALSE;

int equal (coordinate tmp1, coordinate tmp2) {
  return ((tmp1.x == tmp2.x) && (tmp1.y == tmp2.y));
}


int zap_world () {
  int i, j;
  for (i = MINX; i < FIELDX; i++)
    for (j = MINY; j < FIELDY; j++)
      world[i][j] = UNCHARTED;
  goal.x = goal.y = -1;
}


void getRobotInfo (int x, int y, int *team_number, int *robot_number) {
  
  int i, j;

  for (i = 0; i < NUM_TEAMS; i++) 
    for (j = 0; j < team[i].team_size; j++) {
      if ((team[i].robot[j].coord.x == x) &&
	  (team[i].robot[j].coord.y == y)) {
	*team_number = i;
	*robot_number = j;
	return;  /* Force a double break. */
      }
    }
}


void init_world (int id, int place_obstacles) {
  
  int i;
  
  if (place_obstacles) 
    for (i = 0; i < FIELDX+FIELDY; i++) {
      if (static_obstacles)
	world[random()%FIELDX][random()%FIELDY] = STATIC_OBS;
      /*world[random()%FIELDX][random()%FIELDY] = NEG_OBS;*/
      world[random()%FIELDX][random()%FIELDY] = POS_OBS;
    }

  do {
    goal.x = random()%FIELDX;
    goal.y = random()%FIELDY;
  }
  while (world[goal.x][goal.y] != UNCHARTED);

  for (i = 0; i < team[id].team_size; i++) {
    /* Make sure we don't land on another robot, or other object. */
    do {
      team[id].robot[i].coord.x      = random()%FIELDX;
      team[id].robot[i].coord.y      = random()%FIELDY;
      team[id].robot[i].orientation  = random()%FULL_ROTATION;
    }
    while ((world [team[id].robot[i].coord.x][team[id].robot[i].coord.y] != 
	    UNCHARTED) || (equal (team[id].robot[i].coord, goal)));
    
    world [team[id].robot[i].coord.x]
      [team[id].robot[i].coord.y] = GENERIC_ROBOT;
  }
}


void initialize () {
  int i, j;

  for (i = 0; i < NUM_TEAMS; i++)
    for (j = 0; j < MAX_ROBOTS; j++) {
      team[i].registered = FALSE;
      team[i].robot[j].coord.x     = -1;
      team[i].robot[j].coord.y     = -1;
      team[i].robot[j].killed_neg  = FALSE;
      team[i].robot[j].life_points = INIT_LIFE;
    }

  zap_world ();
  num_robots = 0;
  place_obstacles = TRUE;
  registered_clients = 0;
}


int isUserOn (char *name, char *host) {
  int i;

  for(i = 0; i < NUM_TEAMS; i++) {
    if (strcmp (team[i].username, name))
      continue;
    if (strcmp (team[i].hostname, host))
      continue;
    if (team[i].registered)
      return(TRUE); /* We have a match of both host and username. */
  }

  return (FALSE);   /* No match found. */
}


void doNegotiate (textmsg_t *rtext) {
  static int responses = 0;
  int i;

  if (++responses == num_clients) {
    stext.m    = OUTSWAP (text);
    sdata.id   = OUTSWAP (ADMIND_ID);
    stext.type = OUTSWAP (NEGOTIATE);

    for (i = 0; i < num_clients; i++) {
      if (Net_Send (team[i].conn_id, &stext, sizeof (stext)) != 
	  NET_SUCCESS) {
	fprintf (stderr, 
		 "%s: Failure to send to host with connection id:%d\n",
		 progname,
		 team[i].conn_id);
	exit (1);
      }   
    }
  }
}


/* Deal with signon from a client. */
void doSignon (textmsg_t *rtext) {

  int id;
  int c, i, j, k;
  int team_number, robot_number;
  coordinate coord;
  int num_bytes;
  dlock_t lock[FIELDX][FIELDY];

  stext.m    = OUTSWAP (text);
  sdata.id   = OUTSWAP (ADMIND_ID);
  stext.type = OUTSWAP (ACCEPT);
  
  if (registered_clients % num_clients == 0)
    initialize ();

  id = INSWAP (rtext->id);
  if (id == -1) {
    
    /* Find a free ID. */
    for (i = 0; i < num_clients; i++)
      if (team[i].registered == FALSE) {
	id = i;
	break;
      }
    
    if (id == -1) {
      stext.type = OUTSWAP (REJECT);
      strcpy (stext.info.reject.text.text_info, 
	      "Too many clients already registered.");
    }

    if (isUserOn (rtext->info.signon.username, 
		  rtext->info.signon.hostname)) {
      stext.type = OUTSWAP (REJECT);
      strcpy (stext.info.reject.text.text_info, 
	      "User (client) already signed on");
    }
    
    if (stext.type == OUTSWAP (REJECT)) {
      if (Net_Send (conn_id, &stext, sizeof (stext)) != NET_SUCCESS) {
	fprintf (stderr, 
		 "%s: Failure to send to host with connection id:%d\n",
		 progname,
		 conn_id);
	exit (1);
      }
      printf ("Sent reject message:\"%s\"\n", 
	      stext.info.reject.text.text_info);
      return;
    }
  }

  registered_clients++;

  team[id].port        = INSWAP (rtext->info.signon.port);
  team[id].team_number = id;
  team[id].conn_id     = conn_id;
  team[id].registered  = TRUE;
  team[id].team_size   = INSWAP (rtext->info.signon.team_size);
  num_robots          += team[id].team_size;

  strncpy (team[id].hostname, rtext->info.signon.hostname, HOST_LEN);
  team[id].hostname[HOST_LEN-1] = '\0';
  strncpy (team[id].username, rtext->info.signon.username, NAME_LEN);
  team[id].username[NAME_LEN-1] = '\0';
  init_world (id, place_obstacles);
  
  /* Don't place obstacles every time a new user signs on. */
  place_obstacles = FALSE;   
  
  /* If "num_clients" have now signed-on, broadcast signon acceptance
     messages to all clients followed by the initial view of the 
     (battlefield) world. */

  if (registered_clients == num_clients) {

    goal_score = num_robots;   /* Initialize score for reaching goal. */

    for (i = 0; i < num_clients; i++) {
      stext.info.accept.id = OUTSWAP (team[i].team_number);
      stext.info.accept.goal_score  = OUTSWAP (goal_score);
      stext.info.accept.num_clients = OUTSWAP (num_clients);
      for (j = 0; j < num_clients; j++) {
	stext.info.accept.client[j].port = OUTSWAP (team[j].port);
	stext.info.accept.client[j].registered = OUTSWAP (TRUE);
	strcpy (stext.info.accept.client[j].username, team[j].username);
	strcpy (stext.info.accept.client[j].hostname, team[j].hostname);
      }
      stext.info.accept.cp = OUTSWAP (con_protocol);

      if (Net_Send (team[i].conn_id, &stext, sizeof (stext)) != 
	  NET_SUCCESS) {
	fprintf (stderr, 
		 "%s: Failure to send to host with connection id:%d\n",
		 progname,
		 team[i].conn_id);
	exit (1);
      }   
    }

    c = 0;
    for (j = MINY; j < FIELDY; j++)
      for (i = MINX; i < FIELDX; i++) {
      
	/* Setup locks for distributed lock manager. */
	lock[i][j].lock_id = OUTSWAP (c);
	lock[i][j].mgr = OUTSWAP ((c++) % num_clients);

	for (k = 0; k < num_clients; k++) {
	  coord.x = i;
	  coord.y = j;
	  sdata.m = OUTSWAP (setup);
	  sdata.id = OUTSWAP (ADMIND_ID);
	  sdata.ts   = OUTSWAP (0);
	  sdata.size = OUTSWAP (0);	/* No variable-length data. */
	  sdata.position.x           = OUTSWAP (i);
	  sdata.position.y           = OUTSWAP (j);
            
	  if (world[i][j] == GENERIC_ROBOT) {
	    getRobotInfo (i, j, &team_number, &robot_number);

	    printf ("(%d,%d) team number:%d, robot number:%d\n",
		    i, j, team_number, robot_number);

	    if (k == team_number)
	      sdata.position.object_type = OUTSWAP (FRIENDLY_ROBOT);
	    else
	      sdata.position.object_type = OUTSWAP (ENEMY_ROBOT);

	    sdata.position.orientation = 
	      OUTSWAP (team[team_number].robot[robot_number].orientation);
	    sdata.id                 = OUTSWAP (team_number);
	    sdata.rinfo.team_size    = OUTSWAP (team[team_number].team_size);
	    sdata.rinfo.robot_number = OUTSWAP (robot_number);
	    sdata.rinfo.robot_life = OUTSWAP (INIT_LIFE);
	    sdata.rinfo.killed_neg = OUTSWAP (FALSE);
	  }
	  else if (equal (goal, coord)) { 
	    /* Inform clients of goal location. */
	    sdata.position.object_type = OUTSWAP (GOAL);
	    sdata.position.orientation = OUTSWAP (0); /* Don't-care value. */
	  }
	  else {
	    sdata.position.object_type = OUTSWAP (world[i][j]);
	    sdata.position.orientation = OUTSWAP (0); /* Don't-care value. */
	  }
	  sdata.size = OUTSWAP (sizeof (dlock_t)); /* Send lockmgr info. */

	  bcopy (&sdata, msg_buf, sizeof (datamsg_t));
	  bcopy (&lock[i][j], 
		 msg_buf + sizeof (datamsg_t), 
		 sizeof (dlock_t));
	  
	  if (Net_Send (team[k].conn_id, 
			msg_buf, 
			sizeof (sdata) + sizeof (dlock_t)) != 
	      NET_SUCCESS) {
	    fprintf (stderr, 
		     "%s: Failure to send to host with connection id:%d\n",
		     progname,
		     team[k].conn_id);
	    exit (1);
	  }   
	}
      }

    /* Send an ID of -1 to each client to say that data has been 
       broadcast to everyone. */
    for (k = 0; k < num_clients; k++) {
      sdata.m  = OUTSWAP (data);
      sdata.id = OUTSWAP (-1);
      if (Net_Send (team[k].conn_id, &sdata, sizeof (sdata)) != 
	  NET_SUCCESS) {
	fprintf (stderr, 
		 "%s: Failure to send to host with connection id:%d\n",
		 progname,
		 team[k].conn_id);
	exit (1);
      }   
    }
  }
}


/* Deal with signoff from a client. */
void doSignoff (textmsg_t *rtext) {
  
  int i, id, flag;

  if (registered_clients == 0) {
    printf ("All clients now signed off.\n");
    return;
  }

  id = INSWAP (rtext->id);
  printf ("%s: Signoff requested from client %d.\n", 
	  progname, id);
  printf ("Reason: %s\n", rtext->info.signoff.reason);
  flag = INSWAP (rtext->info.signoff.flag);

  team[id].registered = FALSE;
  registered_clients--;

  if (registered_clients) {
    stext.m    = OUTSWAP (text);
    stext.id   = OUTSWAP (id);
    stext.type = OUTSWAP (SIGNOFF);
    strcpy (stext.info.signoff.reason, rtext->info.signoff.reason);
    
    for (i = 0; i < num_clients; i++) {
      if (team[i].registered == FALSE)
	continue;
      if (Net_Send (team[i].conn_id, &stext, sizeof (stext)) != 
	  NET_SUCCESS) {
	fprintf (stderr, 
		 "%s: (doSignoff) Failure to send to host "
		 "with connection id:%d\n",
		 progname,
		 team[i].conn_id);
	exit (1);
      }
      team[i].registered = FALSE;
      registered_clients--;
    }
  }
}


void parse_args (int argc, char *argv[]) {
  int i;
  char *p;
  char s[10];
  con_protocol = BSYNC; /* By default. */
  
  for (i = 1; i < argc; i++) {
    p = argv[i]; 
    if (*p++ == '-') {
      switch (*p++) {

      case 'c':
	if (*p =='\0')
	  p = argv[++i];
	if (*p == 'm')
	  con_protocol = MSYNC;
	else if (*p == 'b')
	  con_protocol = BSYNC;
	else if (*p == 'e')
	  con_protocol = ENTRY;
	break;

      case 'n':                   /* Specify the number of clients. */
	if(*p == '\0')
	  num_clients = atoi (argv[++i]);
	else
	  num_clients = atoi (p);
	break;

      case 'o':
	static_obstacles = TRUE;
	break;

      default:
	fprintf (stderr, 
		 "Usage: %s [-n num_clients [-o][-c[m|b|e]]].\n", 
		 argv[0]);
	break;
      }
    }
  }
  if ((argc == 1) || (num_clients > NUM_TEAMS)) /* Setup a default value. */
    num_clients = 2;
  
  printf ("num_clients=%d\n", num_clients);
  
  switch (con_protocol) {
  case BSYNC:
    strcpy (s, "BSYNC");
    break;
  case MSYNC:
    strcpy (s, "MSYNC");
    break;
  case ENTRY:
    strcpy (s, "entry");
    break;
  default:
    break;
  }
  printf ("Consistency protocol = %s\n", s); 

}


main (int argc, char *argv[]) {

  msg_t msg_type;

  progname = argv[0];

  parse_args (argc, argv);	/* Parse command-line args. */

  initialize ();

  /* TCP/IP initialization. */
  if (Net_Initialize (SERVER_PORT) != NET_SUCCESS) {
    fprintf (stderr, "%s: Failed to initialize port %d\n", 
	     argv[0],
	     SERVER_PORT);
    exit (1);
  }

  while (TRUE) {
    /* Wait for text message from client. */
    Net_Receive (&rtext, &msg_len, &conn_id);
    msg_type = INSWAP (rtext.m);
    
    if (msg_type == text) {
      switch (INSWAP (rtext.type)) {
      case SIGNON:
	printf ("%s(signon).\n", progname);
	printf ("host %s: port %d.\n", 
		rtext.info.signon.hostname, 
		INSWAP (rtext.info.signon.port));
	printf ("user %s.\n", 
		rtext.info.signon.username);
	printf ("team_size %d.\n", 
		INSWAP (rtext.info.signon.team_size));
	doSignon (&rtext);
	break;
	
      case SIGNOFF:
	doSignoff (&rtext);
	break;

      case NEGOTIATE:
	doNegotiate (&rtext);
	break;
	
      case DATAREQ: 
	break;
	
      default:
	fprintf (stderr, "Unexpected message type %d.\n", 
		 INSWAP (rtext.type));
	break;
      }
    }
    else {
      fprintf (stderr, "%s: Expected text message but "
	       "got something else.\n", progname);
      exit(1);
    }
  }
}
