/* Synchronous sound generation. */

#include <unistd.h>
#include <stdio.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>

#ifdef LINUX
#include <linux/soundcard.h>
#endif

#define SOUNDDIR "./sounds"

#define SBUF_SIZE   32
#define BUFFER_SIZE (1024 * SBUF_SIZE)
#define NUM_SOUNDS 8

static char *audio_dev = "/dev/audio";
typedef struct {
  unsigned char buf[BUFFER_SIZE];
  off_t fsize;
} audio_t;

static audio_t audio[NUM_SOUNDS];       
static int audio_fd;

const char *soundfile[NUM_SOUNDS] = {
  "missile",
  "explosion",
  "bomb",
  "bonus",
  "sector",
  "goal",
  "walk",
  "rotate"
};


/* Rather than piping sound to a child to allow asynchrony of parent
   i.e. the parent can continue instead of blocking waiting for sound
   to be piped to the audio device, just write sound stream to the
   sound device in the same process. This will momentarily block the
   process but it will mean the sound is synchronized with the 
   corresponding event in the same process. To cut down on time,
   setUpAudioSystem () preallocates buffers for the soundfiles and 
   opens the audio device. */

void setUpAudioSystem () {
  int i, ifd;
  char filename[256];
  off_t fsize;
  struct stat fbuf;

  if ((audio_fd = open (audio_dev, O_WRONLY)) < 0) {
    fprintf (stderr, "Cannot open audio device.\n");
    return;
  }
  
  for (i = 0; i < NUM_SOUNDS; i++) {
    /* Must be a sound file name. */
    sprintf (filename, "%s/%s.au", SOUNDDIR, soundfile[i]);    
      
    /* Open the sound file for reading. */
    if ((ifd = open (filename, O_RDONLY)) < 0) {
      fprintf (stderr, "Unable to open sound file %s.\n", filename);
      return;
    }
    fstat (ifd, &fbuf);
    fsize = fbuf.st_size;
    audio[i].fsize = fsize;
    if (read (ifd, (char *) audio[i].buf, BUFFER_SIZE) != fsize)
      fprintf (stderr, "Failure to read correct number of "
	       "bytes from sound file.\n");
    close (ifd);
  }
}


void freeAudioSystem () {
  close (audio_fd);
}


void playSoundFile (char *filename) {
  int i;

  for (i = 0; i < NUM_SOUNDS; i++) 
    if (strcmp (filename, soundfile[i]) == 0)
      break;

  /* At this point, we're all ready to copy the data. */
  write (audio_fd, (char *) audio[i].buf, audio[i].fsize);

  /* Flush any audio activity. Right now, I only know 
     how to do this for LINUX. */

#ifdef LINUX

  if (ioctl (audio_fd, SNDCTL_DSP_SYNC, 0) < 0)
    fprintf (stderr, "Unable to flush audio device.\n");

#endif
};


/* Int Main () { */
/*   Char Soundfile[64]; */
  
/*   Setupaudiosystem (); */
/*   While (1) { */
/*     Printf ("Enter Sound File:\N"); */
/*     Scanf ("%S", Soundfile); */
/*     If (Strcmp (Soundfile, "Quit") == 0) */
/*       Break; */
/*     Playsoundfile (Soundfile); */
/*   } */

/*   Freeaudiosystem (); */
  
/*   Return 0; */
/* } */







