/* General consistency functions. */

#include "consistency.h"

extern int num_clients;
extern unsigned long current_time;

void initLocks () {

  int i, j;
  
  for (i = 0; i < FIELDX; i++)
    for (j = 0; j < FIELDY; j++) {
      mgr_lock[i][j].mgr             = -1;
      mgr_lock[i][j].lock_id         = -1;
      mgr_lock[i][j].modified        = FALSE;
      mgr_lock[i][j].type            = LK_WRITE;
      mgr_lock[i][j].ts              = 0;
      mgr_lock[i][j].client_id       = -1;
      mgr_lock[i][j].pending_writers = NULL;
      
      client_lock[i][j].mgr             = -1;
      client_lock[i][j].lock_id         = -1;
      client_lock[i][j].modified        = FALSE;
      client_lock[i][j].type            = -1;
      client_lock[i][j].ts              = 0;
      /* If status is 'locked' the local client holds a write lock. */
      client_lock[i][j].status          = unlocked;
      client_lock[i][j].pending_writers = NULL;
    }
  lock_granted = FALSE;
}


void setupConProto (int protocol) {
  
  int i, j;
  int min_dist [NUM_TEAMS];
  sched_t exchange[NUM_TEAMS];

  switch (protocol) {
  case MSYNC:
  case BSYNC:
    el_front = NULL;

    for (i = 0; i < NUM_TEAMS; i++) {
      sync_list[i].ts          = 0;
      sync_list[i].pending     = 0;
      sync_list[i].sync_flag   = FALSE;
      sync_list[i].resync_flag = TRUE;
      for (j = 0; j < MAX_ROBOTS; j++)
	event_buffer[i][j] = FALSE;
    }

    sortExchangeTimes (calcShortestDist 
		       (ALL_CLIENTS, min_dist), 
		       exchange,
		       num_clients);

    for (i = 0; i < num_clients; i++) {
      if (con_protocol == MSYNC) 
	addExchangeTimeToList (exchange[i].client_id, 
			       exchange[i].time,
			       &el_front);
      else if (con_protocol == BSYNC)
	addExchangeTimeToList (i, 
			       current_time + 1,
			       &el_front);
    }
    break;

  case ENTRY:
    for (i = 0; i < FIELDX; i++)
      for (j = 0; j < FIELDY; j++)
	/* NOTE: Must call this function AFTER this client
	   has a valid 'id'. */
	client_lock[i][j].client_id = id;
    break;
    
  default:
    break;
    
    /* Add other consistency protocol initializations here. */
  }
}





