/* Look-Ahead Consistency. */

#include "consistency.h"

extern unsigned long current_time;
extern int num_clients;
extern int team_size;
extern client_t client [NUM_TEAMS];
extern int free_data_token;

/* Calculate shortest distance from a robot in this client team
   to the nearest robot in each of the chosen other teams. */
int *calcShortestDist (int client_number,
		       int *min_dist) {
  
  int i, j, k, l, x, y, end;
  int worst_dist, x_dist, y_dist, min_x_dist, min_y_dist;

  worst_dist = (FIELDX > FIELDY) ? FIELDX : FIELDY;
  
  if (client_number == ALL_CLIENTS)
    for (i = 0; i < num_clients; i++) {
      min_dist [i] = worst_dist;
    }
  else
    min_dist [client_number] = worst_dist;

  for (i = 0; i < team_size; i++) {
    x = robot_position [id][i].x;
    y = robot_position [id][i].y;
    
    if ((x == -1) || (y == -1) ||
	((x == goal.x) && (y == goal.y)) ||
	(robot_life[id][i] <= 0))
      continue;
    
    if (client_number == ALL_CLIENTS) {
      j = 1;
      end = num_clients;
    }
    else {
      j = client_number;
      end = client_number+1;
    }

    for (; j < end; j++) {
      min_x_dist = min_y_dist = worst_dist;
      k = (client_number == ALL_CLIENTS) ? 
	(id+j) % num_clients : j;
      for (l = 0; l < client[k].team_size; l++) {
	if ((robot_position[k][l].x == -1) || 
	    (robot_position[k][l].y == -1) ||
	    ((robot_position[k][l].x == goal.x) && 
	    (robot_position[k][l].y == goal.y)) ||
	    (robot_life[k][l] <= 0))
	  continue;

	x_dist = abs (x - robot_position[k][l].x);
	y_dist = abs (y - robot_position[k][l].y);
	
	min_x_dist = (min_x_dist < x_dist) ?
	  min_x_dist : x_dist;
	min_y_dist = (min_y_dist < y_dist) ?
	  min_y_dist : y_dist;

	min_dist [k] = (min_dist [k] < x_dist) ?
	  min_dist [k] : x_dist;
	min_dist [k] = (min_dist [k] < y_dist) ?
	  min_dist [k] : y_dist;
      }

      if (((min_x_dist == 1) && (min_y_dist == 1)) ||
	  ((min_x_dist == 2) && (min_y_dist == 0)) || 
	  ((min_x_dist == 0) && (min_y_dist == 2))) {
	if (k < id)
	  free_data_token = FALSE;
	min_dist [k] = 0;
      }
      /* Added the following to deal with MAX_RANGE. */
      else {
	min_dist [k] = (((min_x_dist + min_y_dist + 1) - MAX_RANGE) > 0) ?
	  (min_x_dist + min_y_dist + 1) - MAX_RANGE : 0;
      }
    }
  }
  
  return (min_dist);
}


static int compare (sched_t *a, 
		    sched_t *b) {

  return (a->time - b->time);
}


void sortExchangeTimes (int *min_dist,
			sched_t *exchange,
			int num) {
  
  int i;

  for (i = 0; i < num; i++) {
    exchange[i].time = current_time + floor ((double)min_dist[i]/2);
    exchange[i].client_id = i;
    exchange[i].next = NULL;
  } 

  qsort (exchange, num, sizeof (sched_t), compare);
  
}


/* Build linked-list of ordered exchange times with each client. */
void addExchangeTimeToList (int client_id,
		      int exchange_time,
		      sched_t **front) {

  sched_t *tmp, *tmp2, *prev;
  int bogus_time;
  
  bogus_time = current_time + 
    floor ((double)((FIELDX > FIELDY) ? FIELDX : FIELDY)/2);

  if (exchange_time == bogus_time)
    return;

  if (*front == NULL) {
    *front = (sched_p) malloc (sizeof (sched_t));
    (*front)->time = exchange_time;
    (*front)->client_id = client_id;
    (*front)->next = NULL;
  }
  else {
    tmp = (sched_p) malloc (sizeof (sched_t));
    tmp->time = exchange_time;
    tmp->client_id = client_id;
    tmp2 = *front;

    while (tmp2 != NULL) {
      if (tmp->time > tmp2->time) {
	prev = tmp2;
	tmp2 = tmp2->next;
      }
      else
	break;
    }
    
    if (tmp2 == *front)
      *front = tmp;
    else
      prev->next = tmp;
    tmp->next = tmp2;
  }
  
}


void delExchangeTimeFromList (int client_id,
			      sched_t **front) {
  
  sched_t *ptr, *prev;
  
  ptr = prev = *front;

  while (ptr != NULL) {
    if (ptr->client_id == client_id) {
      if (ptr == *front)
	*front = ptr->next;
      else
	prev->next = ptr->next;

      free (ptr);
      break;
    }
    else {
      prev = ptr;
      ptr = ptr->next;
    }
  }
}







