
proc MenuSetup { {frame .menubar} } {
    global Menu
    frame $frame -bd 3 -relief raised -bg AntiqueWhite
    pack $frame -side top -fill x
    set Menu(menubar) $frame
    set Menu(uid) 0
}


proc Menu { label {side "left"} {underline {}} } {
    global Menu left_right
    if [info exists Menu(menu,$label)] {
	error "Menu $label already defined"
    }
    set name $Menu(menubar).mb$Menu(uid)
    set menuName $name.menu
    incr Menu(uid)
    set mb [menubutton $name -text $label -bg AntiqueWhite \
	    -underline $underline -menu $menuName]
    if {$side == "left"} {
	
	pack $mb -side left -ipady 5 -padx 4 -pady 2
    } else {
	pack $mb -side right -ipady 5 -padx 4 -pady 2
    }
    set menu [menu $menuName -tearoff 1 -bg AntiqueWhite]
    
    # Remember the widget name under a variable derived from the label.
    # This allows MenuBind to be passed the label instead of the widget.
    set Menu(menu,$label) $menu
    return $menu
}


proc UpdateSectorInfo { sector } {
    global Menu

    if [winfo exists $Menu(menubar).im1] {
	destroy $Menu(menubar).im1
    }
    if [winfo exists $Menu(menubar).im2] {
	destroy $Menu(menubar).im2
    }
    if [winfo exists $Menu(menubar).message] {
	destroy $Menu(menubar).message
    }
	
    switch $sector {
	"Top Left" { 
	    set im1 [image create bitmap -file bitmaps/a_up.xbm]
	    set im2 [image create bitmap -file bitmaps/a_left.xbm]
	    set text "Fields of War"
	}
	"Top Right" { 
	    set im1 [image create bitmap -file bitmaps/a_up.xbm]
	    set im2 [image create bitmap -file bitmaps/a_right.xbm]
	    set text "Desert Storm"
	}
	"Bottom Left" { 
	    set im1 [image create bitmap -file bitmaps/a_down.xbm]
	    set im2 [image create bitmap -file bitmaps/a_left.xbm]
	    set text "Red Dawn"
	}
	"Bottom Right" { 
	    set im1 [image create bitmap -file bitmaps/a_down.xbm]
	    set im2 [image create bitmap -file bitmaps/a_right.xbm]
	    set text "Impact Zone"
	}
    }

    label $Menu(menubar).im1 -image $im1 -bg green -relief raised
    label $Menu(menubar).im2 -image $im2 -bg green -relief raised
    label $Menu(menubar).message -fg red -bg AntiqueWhite \
	-text $text \
	-font -adobe-times-bold-*-normal-*-14-140-75-75-p-77-iso8859-1
    pack $Menu(menubar).im1 $Menu(menubar).im2 -side left 
    pack $Menu(menubar).message -side left -padx 10

#    update idletasks
}


proc ChangeSector { oldSectorName sectorName } {
    global sector old_sector

    if {$oldSectorName != $sectorName} {
	if ![winfo exists .sector] {
	    set s [frame .sector -bd 5 -relief ridge -bg AntiqueWhite]
	    label $s.text -text {Changing Sector...} -bg AntiqueWhite\
		-font -adobe-times-bold-i-normal--18-180-75-75-p-98-iso8859-1
	    pack $s.text -side top -pady 5 -anchor c
	    label $s.num -text $sectorName -fg red -bg AntiqueWhite\
		-font -adobe-times-bold-i-normal--24-240-75-75-p-128-iso8859-1
	    pack $s.num -side top -pady 5 -anchor c
	    place .sector -in . -relx 0.5 -rely 0.5 -x 5 -y 5 -anchor c
	    after 750 destroy .sector
#	    update idletasks
	}
    }
    set sector $sectorName
    set old_sector $sectorName
    UpdateSectorInfo $sector
}


proc MenuCommand { menuName label command } {
    global Menu
    if [catch {set Menu(menu,$menuName)} menu] {
	error "No such menu: $menuName"
    }
    $menu add command -label $label -command $command
}


proc MenuCheck { menuName label var { command  {} } } {
    global Menu
    if [catch {set Menu(menu,$menuName)} menu] {
	error "No such menu: $menuName"
    }
    $menu add check -label $label -command $command \
	    -variable $var
}


proc MenuRadio { menuName label var {value {}} {command {}} } {
    global Menu 
    if [catch {set Menu(menu,$menuName)} menu] {
	error "No such menu: $menuName"
    }
    if {[string length $value] == 0} {
	set value $label
    }
    $menu add radio -label $label -command $command \
	    -value $value -variable $var
}


proc MenuSeparator { menuName } {
    global Menu
    if [catch {set Menu(menu,$menuName)} menu] {
	error "No such menu: $menuName"
    }
    $menu add separator
}

proc MenuCascade { menuName label } {
    global Menu
    if [catch {set Menu(menu,$menuName)} menu] {
	error "No such menu: $menuName"
    }
    if [info exists Menu(menu,$label)] {
	error "Menu $label already defined"
    }
    set sub $menu.sub$Menu(uid)
    incr Menu(uid)
    menu $sub -tearoff 0 -bg AntiqueWhite
    $menu add cascade -label $label -menu $sub
    set Menu(menu,$label) $sub 
}


proc MenuBind { what sequence menuName label } {
    global Menu
    if [catch {set Menu(menu,$menuName)} menu] {
	error "No such menu: $menuName"
    }
    if [catch {$menu index $label} index] {
	error "$label not in menu $menuName"
    }
    set command [$menu entrycget $index -command]
    bind $what $sequence $command
    $menu entryconfigure $index -accelerator $sequence
}


proc MenuEntryConfigure { menuName label args } {
    global Menu
    if [catch {set Menu(menu,$menuName)} menu] {
	error "No such menu: $menuName"
    }
    eval {$menu entryconfigure $label} $args
}


proc AnimationRate {} {
    global anim_delay

    catch {destroy .anim}
    toplevel .anim -background AntiqueWhite
    wm title .anim "Animation Rate"
    wm iconname .anim "Animation"
    wm iconbitmap .anim @bitmaps/icon_pixmap.xbm


    image create photo im -file bitmaps/speed.gif -format gif
    button .anim.image -image im -relief raised \
	-bd 2 -command {destroy .anim}
    pack .anim.image -side top -pady 2
    scale .anim.scale -from 1 -to 1000 -length 150 -variable anim_delay \
	-orient horizontal -label "Delay (mS)" -resolution 1 \
	-troughcolor "light steel blue" -bg AntiqueWhite \
	-relief raised -bd 2 -length 300
    pack .anim.scale
	
    wm withdraw .anim
#    update idletasks
    set x [expr [winfo rootx .] + \
	       ([winfo width .]-[winfo reqwidth .anim])/2]
    set y [expr [winfo rooty .] + \
	       ([winfo height .]-[winfo reqheight .anim])/2]
    
    wm geometry .anim +$x+$y
    wm deiconify .anim
}


proc TitleBar {} {
    global Menu 

    set f [frame $Menu(menubar).f -bg AntiqueWhite -bd 2 -relief sunken]

    label $f.title -text {Battle Robots} -bg AntiqueWhite \
	    -font -adobe-times-bold-i-normal--24-240-75-75-p-128-iso8859-1

    image create photo tank_im -palette 16
    tank_im read bitmaps/tankr.gif -format gif 
    label $f.tank_im -image tank_im -bg AntiqueWhite
    
    pack $f.title $f.tank_im -side left -padx 10
    pack $f -side right
}


proc AboutMessage {} {
    catch {destroy .about}
    toplevel .about -bg AntiqueWhite
    wm title .about "About Battle Robots"
    wm iconname .about "AboutBattle"
    wm iconbitmap .about @bitmaps/icon_pixmap.xbm

    label .about.title -text {Battle Robots} -bg AntiqueWhite \
	    -font -adobe-times-bold-i-normal--24-240-75-75-p-128-iso8859-1
    pack .about.title -side top -pady 15
    image create photo im -file bitmaps/robots.gif -format gif -palette 16
    button .about.image -image im -relief raised \
	-bg AntiqueWhite -command {destroy .about}
    pack .about.image -side top -padx 20 
    message .about.subtitle -width 10c -justify center \
	    -font -adobe-times-bold-i-normal-*-14-140-75-75-p-77-iso8859-1 \
	    -text "Let the games begin..." -bg AntiqueWhite
    pack .about.subtitle -side top -pady 10 -padx 15
    image create photo copyright -file bitmaps/copyright.gif \
	    -format gif -palette 16
    frame .about.text -relief groove -bd 5
    label .about.text.copyright -image copyright
    message .about.text.msg -width 10c \
	-text "1995 By Rich West Georgia Institute of Technology" \
	-font -adobe-times-medium-r-normal-*-12-120-75-75-p-64-iso8859-1
    pack .about.text.copyright -side left
    pack .about.text.msg -side left
    pack .about.text
    wm withdraw .about
#    update idletasks
    set x [expr [winfo rootx .] + \
	       ([winfo width .]-[winfo reqwidth .about])/2]
    set y [expr [winfo rooty .] + \
	       ([winfo height .]-[winfo reqheight .about])/2]
    wm geometry .about +$x+$y
    wm deiconify .about
}


proc MenuConfig {} {
    global sector old_sector \
	opmode end_prog sound

    MenuSetup

    Menu Options left 0
    MenuCommand Options Animation AnimationRate 
    MenuSeparator Options
    MenuCascade Options Sound
    MenuRadio Sound On sound 1
    MenuRadio Sound Off sound 0

    MenuSeparator Options
    MenuCascade Options Mode
    MenuRadio Mode User opmode 
    MenuRadio Mode Computer opmode

    Menu Window left 0
    MenuRadio Window {Top Left} sector {} \
	    {ChangeSector $old_sector {Top Left}}
    MenuRadio Window {Top Right} sector {} \
	    {ChangeSector $old_sector {Top Right}}
    MenuRadio Window {Bottom Left} sector {} \
	    {ChangeSector $old_sector {Bottom Left}}
    MenuRadio Window {Bottom Right} sector {} \
	    {ChangeSector $old_sector {Bottom Right}}

    Menu Quit left 0
    MenuCommand Quit Quit {set end_prog 1}

    Menu Help right 0
    MenuCommand Help About... AboutMessage
}


proc CheckInput {} {
    global dialog

# Set max munber of robots per team at 9 for now.

    if [string match {[1-9]} $dialog(result)] {
	set dialog(ok) 1
    }
}


proc MessageBox { { msg {Waiting For Other Clients} } } {
    if ![winfo exists .mbox] {
	set m [frame .mbox -bd 8 -relief ridge -bg AntiqueWhite]
	label $m.msg -text $msg -bg AntiqueWhite -relief groove -bd 5 \
	    -font -adobe-times-bold-r-normal--24-240-75-75-p-128-iso8859-1
	pack $m.msg -side top -ipadx 5 -ipady 15 -fill x 
	place .mbox -in . -relx 0.5 -rely 0.5 -x 5 -y 5 -anchor c
    }
#    update idletasks
}


proc DialogBox { { prompt {Number of robots in team [1-9]?} } } {
    global dialog team_size

    set fn1 "-adobe-helvetica-bold-r-*-*-14-*-*-*-*-*-*-*"

    if ![winfo exists .dialog] {
	set d [frame .dialog -bd 5 -relief ridge -bg AntiqueWhite]
	message $d.msg -text $prompt -anchor w -relief groove \
	    -width 280 -bg AntiqueWhite
	entry $d.entry -textvariable dialog(result) \
	    -width 4 -bg "light steel blue" -fg red -bd 4 \
	    -font $fn1
	set b [frame $d.buttons -bd 10 -bg AntiqueWhite]
	pack $d.msg -side top -fill x
	pack $d.entry -side left -padx 10
	pack $d.buttons -side left
	bind $d.entry <Return> {CheckInput}
	bind $d.entry <Escape> {set dialog(ok) 0}
	button $b.ok -text OK -command {CheckInput} -bg green2 \
	    -activebackground {green1}
	button $b.cancel -text Cancel -command {set dialog(ok) 0} \
	    -bg red2 -activebackground {red1}
	pack $b.ok -side left
	pack $b.cancel -side left -padx 5

#	update idletasks
	place .dialog -in . -relx 0.5 -rely 0.5 -x 5 -y 5 -anchor c
	focus $d.entry
    }

    if {$dialog(ok) == 1} {
	set team_size $dialog(result)
	destroy .dialog
    } elseif {$dialog(ok) == 0} {
	set team_size 0
	destroy .dialog
    }

    set dialog(ok) -1
}


proc MenuBindings {} {
    MenuBind . "<Up><Left>" Window {Top Left}
    MenuBind . "<Up><Right>" Window {Top Right}
    MenuBind . "<Down><Left>" Window {Bottom Left}
    MenuBind . "<Down><Right>" Window {Bottom Right}
    MenuBind . <Control-A> Help About...
    MenuBind . <Control-c> Quit Quit
}


proc StatusWindow { parent args } {
    global current_time

    frame $parent -bg "light steel blue"

    #Create a listbox attached to scrollbars.
    eval {text $parent.list \
	      -yscrollcommand [list $parent.sy set] \
	      -xscrollcommand [list $parent.pad.sx set]} $args

    scrollbar $parent.sy -orient vertical \
	-command [list $parent.list yview] -width 8 \
	-bg AntiqueWhite

    #Create extra frame to hold pad and horizontal scrollbar.
    frame $parent.pad
    scrollbar $parent.pad.sx -orient horizontal \
	-command [list $parent.list xview] -width 8 \
	-bg AntiqueWhite
    
    #Create padding based on the scrollbar's 
    #width and its borders.
    set pad [expr [$parent.sy cget -width] + 2 * \
		 ([$parent.sy cget -bd] + \
		      [$parent.sy cget -highlightthickness])]
    frame $parent.pad.it -width $pad -height $pad -bg AntiqueWhite

    #Insert time-monitor into parent frame.
    label $parent.label -text Time: -relief ridge \
	-font -adobe-helvetica-bold-r-*-*-14-*-*-*-*-*-*-* \
	-bg AntiqueWhite -bd 2
    label $parent.label.tv -textvariable current_time \
	-bg AntiqueWhite -fg red
    pack $parent.label -pady 0 -side bottom -fill x
    pack $parent.label.tv -padx 2 -side right

    #Arrange everything in the parent frame.
    pack $parent.pad -side bottom -fill x
    pack $parent.pad.it -side right
    pack $parent.pad.sx -side bottom -fill x
    pack $parent.sy -side right -fill y
    pack $parent.list -side left -fill both -expand true

    image create photo score_image -file bitmaps/score.gif -format gif
    label $parent.list.im -image score_image -bg "light steel blue"
    pack propagate $parent.list false
    pack $parent.list.im 
    $parent.list insert end "\n\n"
    $parent.list configure -state disabled

#    pack $parent -side right
#    update idletasks
}


proc ClearStatusWindow { parent } {

    $parent.list configure -state normal
    $parent.list delete 1.0 end
    $parent.list insert end "\n\n"
    $parent.list configure -state disabled
#    update idletasks
}


proc InitStatusWindow { parent str } {

    $parent.list configure -state normal
    $parent.list insert end $str
    $parent.list configure -state disabled
#    update idletasks
}


proc RefreshStatusWindow { parent str strcmp } {
    
    $parent.list configure -state normal

    #Delete line in status window which matches strcmp.
    set index [$parent.list search -forw -exact $strcmp 1.0]

    #Delete whole line from the index to the newline character.
    $parent.list delete $index "$index + 1 line - 1 char"

    #Replace deleted line with new text.
    $parent.list insert $index $str
    $parent.list configure -state disabled
#    update idletasks
}


set dialog(ok) -1
set sector {Top Left}
set old_sector {Top Left}
set end_prog 0
set sound 1
set current_time 0

#opmode is either "Computer"(non-interactive) or "User"(interactive).
set opmode Computer
MenuConfig
MenuBindings
UpdateSectorInfo {Top Left}
TitleBar
# StatusWindow .l -bd 1 -relief ridge \
#     -width 20 -height 13 \
#     -bg "light steel blue" \
#     -wrap none \
#     -font -adobe-helvetica-bold-r-*-*-12-*-*-*-*-*-*-*

# InitStatusWindow .l "root@cerberus\t0\t0\n"
# InitStatusWindow .l "root@speedking\t0\t0\n"
# InitStatusWindow .l "root@timewarp\t0\t0\n"
# InitStatusWindow .l "rich@cerberus\t0\t0\n"
# InitStatusWindow .l "rich@speedking\t0\t0\n"
# InitStatusWindow .l "rich@timewarp\t0\t0\n"
# InitStatusWindow .l "john@cerberus\t0\t0\n"
# InitStatusWindow .l "john@speedking\t0\t0\n"
# InitStatusWindow .l "john@timewarp\t0\t0\n"

# RefreshStatusWindow .l "root@timewarp\t50\t150" "root@timewarp"
# RefreshStatusWindow .l "john@speedking\t50\t150" "john@speedking"
# RefreshStatusWindow .l "john@timewarp\t50\t150" "john@timewarp"
# RefreshStatusWindow .l "john@speedking\t50\t250" "john@speedking"
# ClearStatusWindow .l
# InitStatusWindow .l "john@speedking\t0\t0\n"

# for {set i 0} {$i < 10000} {incr i} {
#     incr current_time
#     update
# }











