/* Setup world as seen by each client's (team's) robots . */
#include "consistency.h"

int world [FIELDX][FIELDY];
coordinate goal; /* Same global goal for all robots. */
coordinate environment [MAX_ROBOTS][NUM_OBJECT_TYPES+OBJECT_OFFSET];
int sighted_flag [MAX_ROBOTS][NUM_OBJECT_TYPES+OBJECT_OFFSET];

int robot_front [NUM_TEAMS][MAX_ROBOTS];
int robot_life [NUM_TEAMS][MAX_ROBOTS];
coordinate robot_position [NUM_TEAMS][MAX_ROBOTS];
int killed_neg [NUM_TEAMS][MAX_ROBOTS];

missile_type missile [MAX_MISSILES];

extern int team_size;
extern char *pause_string;
extern dlock_t client_lock[FIELDX][FIELDY];
extern int ECgetData (int client_id, int x, int y);

int equal (coordinate tmp1, coordinate tmp2) {
  return ((tmp1.x == tmp2.x) && (tmp1.y == tmp2.y));
}


int zap_world () {
  int i, j;
  for (i = MINX; i < FIELDX; i++)
    for (j = MINY; j < FIELDY; j++)
      world[i][j] = UNCHARTED;
  goal.x = goal.y = -1;
}


float distance (coordinate c1, coordinate c2) {
  return (sqrt(((c1.x - c2.x)*(c1.x-c2.x)) + 
	       ((c1.y - c2.y)*(c1.y - c2.y))));
}


int in_world (coordinate cur_pos) {
  return ((cur_pos.x >= MINX) && (cur_pos.x < FIELDX)
	  && (cur_pos.y >=MINY) && (cur_pos.y < FIELDY));
}


int already_visited (coordinate c) {
  return (!(world[c.x][c.y] == UNCHARTED));
}


coordinate up_left (coordinate tmp) {
  tmp.x--;
  tmp.y--;
  return tmp;
}


coordinate up (coordinate tmp) {
  tmp.y-- ;
  return tmp;
}


coordinate up_right (coordinate tmp) {
  tmp.x++;
  tmp.y--;
  return tmp;
}


coordinate left (coordinate tmp) {
  tmp.x--;
  return tmp;
}


coordinate right (coordinate tmp) {
  tmp.x++;
  return tmp;
}


coordinate dn_left (coordinate tmp) {
  tmp.x--;
  tmp.y++;
  return tmp;
}


coordinate dn (coordinate tmp) {
  tmp.y++ ;
  return tmp;
}


coordinate dn_right (coordinate tmp) {
  tmp.x++;
  tmp.y++;
  return tmp;
}


void update_scan_buffer (int robot_id,
			 int (*sighted_flag)
			 [NUM_OBJECT_TYPES+OBJECT_OFFSET],
			 coordinate (*direction_func)(coordinate),
			 float *mindist,
			 int missile_orientation,
			 int range,
			 datareq_t (*client_list)[MAX_RANGE]) 
{
  
  int tmp;
  int count[NUM_TEAMS];
  coordinate scan_coordinate;
  int team_number, robot_number;
    
  bzero (count, NUM_TEAMS*sizeof(int));
  
  scan_coordinate = robot_position[id][robot_id];

  for (tmp = 1; ((tmp < range) && (range <= MAX_RANGE)); tmp++) {
    scan_coordinate = direction_func (scan_coordinate);
      
    if (in_world (scan_coordinate)) {

      if (con_protocol == ENTRY) {

	if (client_lock[scan_coordinate.x][scan_coordinate.y].status !=
	    locked) {
	  client_lock[scan_coordinate.x][scan_coordinate.y].type = 
	    (tmp == 1) ? LK_WRITE : LK_READ;

	  /* printf ("Acquiring %s-lock at (%d, %d) from lock-mgr %d\n", 
	     (client_lock[scan_coordinate.x][scan_coordinate.y].type == 
	     LK_WRITE) ? "write" : "read",
	     scan_coordinate.x,
	     scan_coordinate.y,
	     client_lock[scan_coordinate.x][scan_coordinate.y].mgr);*/
	  
	  while (acquireLock (id,
			      &client_lock
			      [scan_coordinate.x][scan_coordinate.y], 
			      scan_coordinate.x, 
			      scan_coordinate.y) != CLIENT_LK_ACQ)
	    while (eventCheck (FALSE, &pause_string));

	  /* printf ("Acquired %s-lock at (%d, %d)\n", 
	     (client_lock[scan_coordinate.x][scan_coordinate.y].type == 
	     LK_WRITE) ? "write" : "read",
	     scan_coordinate.x,
	     scan_coordinate.y);*/
	}
	
	if (client_lock[scan_coordinate.x][scan_coordinate.y].modified) {
	  if (client_lock[scan_coordinate.x][scan_coordinate.y].client_id !=
	      id) {
	    while (ECgetData (client_lock
			      [scan_coordinate.x][scan_coordinate.y].client_id,
			      scan_coordinate.x,
			      scan_coordinate.y) != DATAMSG)
	      while (eventCheck (FALSE, &pause_string));
	  }
	  client_lock[scan_coordinate.x][scan_coordinate.y].modified = FALSE;
	  /* printf ("Got data at (%d, %d)\n", 
	     scan_coordinate.x,
	     scan_coordinate.y);*/
	}
      }

      switch (world[scan_coordinate.x][scan_coordinate.y]) {
      case FRIENDLY_ROBOT :
	if (distance (scan_coordinate, robot_position[id][robot_id]) <
	    mindist [FRIENDLY_ROBOT]) {
	  mindist [FRIENDLY_ROBOT] =
	    distance (scan_coordinate, robot_position[id][robot_id]);
	  sighted_flag [robot_id][FRIENDLY_ROBOT] = TRUE;
	  environment[robot_id][FRIENDLY_ROBOT] = scan_coordinate;
	}
	break;

      case ENEMY_ROBOT :
	if (client_list != NULL) {
	  getRobotInfo (scan_coordinate.x, scan_coordinate.y, 
			&team_number, &robot_number);
	  client_list[team_number][count[team_number]].flag = 
	    ENEMY_ROBOT;
	  client_list[team_number][count[team_number]].x   = 
	    scan_coordinate.x;
	  client_list[team_number][count[team_number]++].y = 
	    scan_coordinate.y;
	}

	if (distance (scan_coordinate, robot_position[id][robot_id]) <
	    mindist [ENEMY_ROBOT]) {
	  mindist [ENEMY_ROBOT] =
	    distance (scan_coordinate, robot_position[id][robot_id]);
	  sighted_flag [robot_id][ENEMY_ROBOT] = TRUE;
	  environment [robot_id][ENEMY_ROBOT] = scan_coordinate;
	}
	break;
	
      case UNKNOWN :
	if (client_list != NULL) {
	  getRobotInfo (scan_coordinate.x, scan_coordinate.y, 
			&team_number, &robot_number);
	  client_list[team_number][count[team_number]].flag = 
	    UNKNOWN;
	  client_list[team_number][count[team_number]].x   = 
	    scan_coordinate.x;
	  client_list[team_number][count[team_number]++].y = 
	    scan_coordinate.y;
	}
	break;
      
      case N_MISSILE :
      case E_MISSILE :
      case S_MISSILE :
      case W_MISSILE :
	if (world[scan_coordinate.x][scan_coordinate.y] ==
	    missile_orientation)
		
	  if (distance (scan_coordinate, robot_position[id][robot_id]) <
	      mindist [MISSILE]) {
	    mindist [MISSILE] = 
	      distance (scan_coordinate, robot_position[id][robot_id]);
	    sighted_flag[robot_id][MISSILE] = TRUE;
	    environment[robot_id][MISSILE] = scan_coordinate;
	  }
	break;

      case POS_OBS :
	if (distance (scan_coordinate, robot_position[id][robot_id]) <
	    mindist [POS_OBS]) {
	  mindist [POS_OBS] = 
	    distance (scan_coordinate, robot_position[id][robot_id]);
	  sighted_flag [robot_id][POS_OBS] = TRUE;
	  environment[robot_id][POS_OBS] = scan_coordinate;
	}
	break;

      case NEG_OBS :
	/*      if(distance(scan_coordinate,robot_position[id][robot_id]) <
		mindist [NEG_OBS]) */
	
	/* Changed so that we only concern ourselves with 
	   bombs in adjacent squares to robot. */

	if (distance (scan_coordinate, robot_position[id][robot_id]) == 1) {
	  mindist [NEG_OBS] = 
	    distance (scan_coordinate, robot_position[id][robot_id]);
	  sighted_flag [robot_id][NEG_OBS] = TRUE;
	  environment[robot_id][NEG_OBS] = scan_coordinate;
	}
	break;

      case STATIC_OBS :
	if (distance (scan_coordinate, robot_position[id][robot_id]) == 1) {
	  mindist [STATIC_OBS] = 
	    distance (scan_coordinate, robot_position[id][robot_id]);
	  sighted_flag [robot_id][STATIC_OBS] = TRUE;
	  environment[robot_id][STATIC_OBS] = scan_coordinate;
	}
	break;

      default :
	break;
	
      }

      if (con_protocol == ENTRY)
	/* If we have a read-locked object at this point it is safe to
	   unlock the object. */
	if (client_lock[scan_coordinate.x][scan_coordinate.y].type == 
	    LK_READ){
	  sendRelLock (id, 
		       &client_lock[scan_coordinate.x][scan_coordinate.y]);
	}
    }
    else
      break;
  }

  /* If we are using entry consistency and need to lock blocks we must 
   lock the robot's block after locking all blocks west (left) of the robot.
   This ensures we lock blocks in increasing id order, so that deadlocks 
   are prevented. */
  if (direction_func == left) 
    if (con_protocol == ENTRY) {
      /* Acquire write lock at location of this robot. This is necessary
	 if the robot moves to another cell or rotates, allowing the lock 
	 manager to be informed that the block where the robot was located 
	 has changed. */

      scan_coordinate = robot_position[id][robot_id];
      if (client_lock[scan_coordinate.x][scan_coordinate.y].status !=
	  locked) {

	/*printf ("Acquiring write-lock at (%d, %d) from lock-mgr %d\n", 
	  scan_coordinate.x,
	  scan_coordinate.y,
	  client_lock[scan_coordinate.x][scan_coordinate.y].mgr);*/
	
	client_lock[scan_coordinate.x][scan_coordinate.y].type = LK_WRITE;
	
	while (acquireLock (id,
			    &client_lock
			    [scan_coordinate.x][scan_coordinate.y], 
			    scan_coordinate.x, 
			    scan_coordinate.y) != CLIENT_LK_ACQ)
	  while (eventCheck (FALSE, &pause_string));

	/*printf ("Acquired write-lock at (%d, %d)\n", 
	  scan_coordinate.x,
	  scan_coordinate.y);*/
      }
    }

  if (mindist [FRIENDLY_ROBOT] < mindist [ENEMY_ROBOT]) 
    
    /* A friendly robot is in line of sight before the sighted enemy. 
       To avoid possibly creating a task to shoot enemy, thereby 
       hitting a team member (another friendly robot), set sighting
       of enemy to false. */
    
    sighted_flag [robot_id][ENEMY_ROBOT] = FALSE;
}


void look (int robot_id, 
	   int (*sighted_flag)[NUM_OBJECT_TYPES+OBJECT_OFFSET],
	   int direction,
	   int range,
	   datareq_t (*client_list)[MAX_RANGE])
{
  int tmp, i, j;
  float mindist[NUM_OBJECT_TYPES+OBJECT_OFFSET];
  
  for (tmp=OBJECT_OFFSET; tmp < NUM_OBJECT_TYPES+OBJECT_OFFSET; tmp++) {
    mindist[tmp] = MAXDIST;
    sighted_flag [robot_id][tmp] = FALSE;
  }

  if (client_list != NULL)
    for (i = 0; i < NUM_TEAMS; i++) 
      for (j = 0; j < MAX_RANGE; j++)
	client_list[i][j].flag = FALSE;

  switch (direction) {
  case NORTH:
    update_scan_buffer (robot_id, sighted_flag, up, 
			mindist, S_MISSILE, range,
			client_list);
    break;
    
  case SOUTH:
    update_scan_buffer (robot_id, sighted_flag, dn, 
			mindist, N_MISSILE, range,
			client_list);
    break;

  case WEST:
    update_scan_buffer (robot_id, sighted_flag, left, 
			mindist, E_MISSILE, range,
			client_list);
    break;
    
  case EAST:
    update_scan_buffer (robot_id, sighted_flag, right, 
			mindist, W_MISSILE, range,
			client_list);
    break;
  }
}
		

int check_location (int direction, int robot_id) {

  coordinate c;
  
  switch (direction) {
  case NORTH:
    c.x = robot_position[id][robot_id].x;
    c.y = robot_position[id][robot_id].y-1;
    if (!in_world(c))
      return NOT_IN_WORLD;
    else
      return (world [robot_position[id][robot_id].x]
	      [robot_position[id][robot_id].y-1]);

  case EAST:
    c.x = robot_position[id][robot_id].x+1;
    c.y = robot_position[id][robot_id].y;
    if (!in_world(c))
      return NOT_IN_WORLD;
    else
      return (world [robot_position[id][robot_id].x+1]
	      [robot_position[id][robot_id].y]);    

  case SOUTH:
    c.x = robot_position[id][robot_id].x;
    c.y = robot_position[id][robot_id].y+1;
    if (!in_world(c))
      return NOT_IN_WORLD;
    else
      return (world [robot_position[id][robot_id].x]
	      [robot_position[id][robot_id].y+1]);    

  case WEST:
    c.x = robot_position[id][robot_id].x-1;
    c.y = robot_position[id][robot_id].y;
    if (!in_world(c))
      return NOT_IN_WORLD;
    else
      return (world [robot_position[id][robot_id].x-1]
	      [robot_position[id][robot_id].y]);    
  }
}


int next_move (coordinate src, 
	       coordinate dest) {

  int xdiff, ydiff;

  xdiff = src.x - dest.x;
  ydiff = src.y - dest.y;
  
  if ((xdiff > 0) && (ydiff > 0))
    return NORTHWEST;

  if ((xdiff > 0) && (ydiff < 0))
    return SOUTHWEST;

  if ((xdiff < 0) && (ydiff > 0))
    return NORTHEAST;

  if ((xdiff < 0) && (ydiff < 0))
    return SOUTHEAST;

  if (xdiff > 0)
    return WEST;

  if (xdiff < 0)
    return EAST;

  if (ydiff > 0)
    return NORTH;

  if (ydiff < 0)
    return SOUTH;
}


int calculate_walk_orientation (int robot_id) {

  int direction;
  int preferred, alternative;
  int i, sign;
  int clear_direction = -1; /* Best direction with no obstacles. */
  int tmp;
  int xdiff, ydiff;

  /* Get preferred direction to goal first. */
  preferred = next_move (robot_position[id][robot_id], goal);

  xdiff = robot_position[id][robot_id].x - goal.x;
  ydiff = robot_position[id][robot_id].y - goal.y;

  switch (preferred) {
  case NORTH:
  case NORTHEAST:
  case NORTHWEST:
    if (preferred == NORTHEAST) {
      if (abs(xdiff) > abs(ydiff)) { 
	preferred   = EAST;
	alternative = NORTH;
      }
      else {
	preferred   = NORTH;
	alternative = EAST;
      }
    }
    else if (preferred == NORTHWEST) {
      if (abs(xdiff) > abs(ydiff)) { 
	preferred   = WEST;
	alternative = NORTH;
      }
      else {
	preferred   = NORTH;
	alternative = WEST;
      }
    }
    else {
      preferred   = NORTH;
      alternative = NORTH;
    }
    break;

  case SOUTH:
  case SOUTHEAST:
  case SOUTHWEST:
    if (preferred == SOUTHEAST) {
      if (abs(xdiff) > abs(ydiff)) { 
	preferred   = EAST;
	alternative = SOUTH;
      }
      else {
	preferred   = SOUTH;
	alternative = EAST;
      }
    }
    else if (preferred == SOUTHWEST) {
      if (abs(xdiff) > abs(ydiff)) { 
	preferred   = WEST;
	alternative = SOUTH;
      }
      else {
	preferred   = SOUTH;
	alternative = WEST;
      }
    }
    else {
      preferred   = SOUTH;
      alternative = SOUTH;
    }
    break;

  case EAST:
    preferred = alternative = EAST;
    break;

  case WEST:
    preferred = alternative = WEST;
    break;
  }

  sign = +1;
  direction = preferred;

  for (i = 1; i < FULL_ROTATION; i++) {
    if (check_location (direction, robot_id) == UNCHARTED) {
      if (clear_direction == -1)
	clear_direction = direction;
    }
    else if (check_location (direction, robot_id) == POS_OBS)
      if (robot_life[id][robot_id] < INIT_LIFE * 2)
	return (direction);

    direction = (direction + FULL_ROTATION + 
		 (sign * (FULL_ROTATION - i))) % FULL_ROTATION;

    sign = (sign == (-1)) ? +1 : -1;
  }
  if ((check_location (preferred, robot_id) == UNCHARTED) ||
      (check_location (preferred, robot_id) == GOAL))
    return (preferred);
  if (check_location (alternative, robot_id) == UNCHARTED)
    return (alternative);
  if (clear_direction != -1)
    return (clear_direction);
  return (preferred);
}
